# This code is provided under a CC-BY-4.0 License
# Published in: Eric D. Taillard
# Design of Heuristic Algorithms for Hard Optimization, Springer Nature, 2023
# https://link.springer.com/book/10.1007/978-3-031-13714-3

from random_generators import unif
from tsp_utilities import tsp_length

######### Building a solution using artificial pheromone trails
def generate_solution_trail(d,                                # Distance matrix
                            tour,                    # Tour produced by the ant
                            trail):                          # Pheromone trails
    n = len(tour)
    for i in range(1, n - 1):
        total = 0
        for j in range(i + 1, n):
            total += trail[tour[i - 1]][tour[j]]
        target = unif(0, total - 1)
        j = i
        total = trail[tour[i - 1]][tour[j]]
        while total < target:
            total += trail[tour[i - 1]][tour[j + 1]]
            j += 1
        tour[j], tour[i] = tour[i], tour[j]
    return tour, tsp_length(d, tour)
