# This code is provided under a CC-BY-4.0 License
# Published in: Eric D. Taillard
# Design of Heuristic Algorithms for Hard Optimization, Springer Nature, 2023
# https://link.springer.com/book/10.1007/978-3-031-13714-3

from build_2opt_data_structure import build_2opt_data_structure  
from tsp_utilities import tsp_2opt_data_structure_to_tour 

######### Local search with 2-opt neighbourhood and first improvement policy
def tsp_2opt_first(d,                                         # Distance matrix
                   tour,                                             # Solution 
                   length):                                   # Solution length

    n = len(tour)
    t = build_2opt_data_structure(tour)
    i = last_i = 0          # i = starting city || last_i = i - a complete tour
    while t[t[i]] >> 1 != last_i:  # Index i has made 1 turn without impovement
        j = t[t[i]]                                
        while j >> 1 != last_i and (t[j]>>1 != last_i or i>>1 != last_i):
            delta =  d[i>>1][j>>1]    + d[t[i]>>1][t[j]>>1] \
                   - d[i>>1][t[i]>>1] - d[j>>1][t[j]>>1]
            if delta < 0:                          # An improving move is found
                next_i, next_j = t[i], t[j]                      # Perform move 
                t[i], t[j] = j ^ 1, i ^ 1               
                t[next_i ^ 1], t[next_j ^ 1] = next_j, next_i

                length += delta                          # Update solution cost
                last_i = i >> 1   # Solution improved: i must make another turn
                j = t[i]    
            j = t[j]                                                   # Next j
        i = t[i]                                                       # Next i

    return tsp_2opt_data_structure_to_tour(t), length
