# This code is provided under a CC-BY-4.0 License
# Published in: Eric D. Taillard
# Design of Heuristic Algorithms for Hard Optimization, Springer Nature, 2023
# https://link.springer.com/book/10.1007/978-3-031-13714-3

from random_generators import unif 
from tsp_utilities import tsp_length
from tsp_LK import tsp_LK 

######### Variable Neighbourhood Search for the TSP
def tsp_VNS(d,                                                # Distance matrix
            best_tour,                                               # TSP tour
            best_length):            
         
    n = len(best_tour)
    iteration, k = 0, 1
    while k < n:
        tour = best_tour[:]
        
        for _ in range(k):                                # Perturbate solution
            u = unif(0, n - 1)
            v = unif(0, n - 1)
            tour[u], tour[v] = tour[v], tour[u]
        length = tsp_length(d, tour)
        tour, length = tsp_LK(d, tour, length)
        iteration += 1
        if length < best_length:
            best_tour = tour[:]
            best_length = length
            print('VNS {:d}\t {:d}\t {:d}'
                  .format(iteration, k, length))
            k = 1
        else:
            k += 1
    return best_tour, best_length
