# This code is provided under a CC-BY-4.0 License
# Published in: Eric D. Taillard
# Design of Heuristic Algorithms for Hard Optimization, Springer Nature, 2023
# https://link.springer.com/book/10.1007/978-3-031-13714-3

from random_generators import unif 
from tsp_utilities import tsp_length 
from tsp_LK import tsp_LK 

######### Record to record iterative local search for the TSP
def tsp_record_to_record(d,                                   # Distance matrix
                         iterations,                     # Number of iterations
                         best_tour,                                  # TSP tour
                         best_length):                  
    n = len(d[0])
    for iteration in range(0, iterations):
        tour = best_tour[:]          # No tolerance: always revert to best tour
        
        for _ in range(2):                                # Perturbate solution
            u = unif(0, n - 1)
            v = unif(0, n - 1)
            tour[u], tour[v] = tour[v], tour[u]
        length = tsp_length(d, tour)
        tour, length = tsp_LK(d, tour, length)
        iteration += 1
        if length < best_length:
            best_tour = tour[:]
            best_length = length
            print('Record to record {:d}\t {:d}'
                  .format(iteration, length))
    return best_tour, best_length
