# This code is provided under a CC-BY-4.0 License
# Published in: Eric D. Taillard
# Design of Heuristic Algorithms for Hard Optimization, Springer Nature, 2023
# https://link.springer.com/book/10.1007/978-3-031-13714-3

######## (Re-)initialize all trails
def init_trail(initial_value,                    # Initial value for all trails
               trail):                                       # Pheromone trails

    n = len(trail[0])
    for i in range(n):
        for j in range(n):
            trail[i][j] = initial_value
    for i in range(n):
        trail[i][i] = 0
    return trail

######### Updating trail values
def update_trail(tour,                      # Last solution generated by an ant
                 global_best,                            # Global best solution
                 exploration,                  # Reinforcement of last solution
                 exploitation,          # Reinforcement of global best solution
                 trail):                                     # Pheromone trails

    if tour == global_best:
        exploration += 1                      # Give more weight to exploration
        trail = init_trail(exploration, trail)
    else:
        for i in tour:
            n = len(trail[0])
            trail[tour[i]][tour[(i + 1) % n]] += exploration
            trail[global_best[i]][global_best[(i + 1) % n]] += exploitation
    return trail, exploration
